@@ -17,26 +17,19 @@ from asyncio import (
     sleep,
 )
 from contextlib import ExitStack, contextmanager
+from ctypes import c_int, c_void_p, pythonapi
 from subprocess import Popen
 from traceback import format_tb
 from typing import (
     Any,
     Callable,
     Coroutine,
-    Dict,
-    FrozenSet,
     Generator,
     Generic,
     Hashable,
     Iterable,
     Iterator,
-    List,
-    Optional,
-    Set,
-    Tuple,
-    Type,
     TypeVar,
-    Union,
     cast,
     overload,
 )
@@ -48,7 +41,9 @@ from prompt_toolkit.cursor_shapes import AnyCursorShapeConfig, to_cursor_shape_c
 from prompt_toolkit.data_structures import Size
 from prompt_toolkit.enums import EditingMode
 from prompt_toolkit.eventloop import (
+    InputHook,
     get_traceback_from_context,
+    new_eventloop_with_inputhook,
     run_in_executor_with_context,
 )
 from prompt_toolkit.eventloop.utils import call_soon_threadsafe
@@ -108,6 +103,21 @@ _SIGWINCH = getattr(signal, "SIGWINCH", None)
 _SIGTSTP = getattr(signal, "SIGTSTP", None)
 
 
+# The following functions are part of the stable ABI since python 3.2
+# See: https://docs.python.org/3/c-api/sys.html#c.PyOS_getsig
+
+# PyOS_sighandler_t PyOS_getsig(int i)
+pythonapi.PyOS_getsig.restype = c_void_p
+pythonapi.PyOS_getsig.argtypes = (c_int,)
+
+# PyOS_sighandler_t PyOS_setsig(int i, PyOS_sighandler_t h)
+pythonapi.PyOS_setsig.restype = c_void_p
+pythonapi.PyOS_setsig.argtypes = (
+    c_int,
+    c_void_p,
+)
+
+
 class Application(Generic[_AppResult]):
     """
     The main Application class!
@@ -279,6 +289,7 @@ class Application(Generic[_AppResult]):
         self._is_running = False
         self.future: Future[_AppResult] | None = None
         self.loop: AbstractEventLoop | None = None
+        self._loop_thread: threading.Thread | None = None
         self.context: contextvars.Context | None = None
 
         #: Quoted insert. This flag is set if we go into quoted insert mode.
@@ -662,8 +673,9 @@ class Application(Generic[_AppResult]):
             # See: https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1553
             handle_sigint = False
 
-        async def _run_async(f: "asyncio.Future[_AppResult]") -> _AppResult:
-            self.context = contextvars.copy_context()
+        async def _run_async(f: asyncio.Future[_AppResult]) -> _AppResult:
+            context = contextvars.copy_context()
+            self.context = context
 
             # Counter for cancelling 'flush' timeouts. Every time when a key is
             # pressed, we start a 'flush' timer for flushing our escape key. But
@@ -707,6 +719,16 @@ class Application(Generic[_AppResult]):
                         flush_task.cancel()
                     flush_task = self.create_background_task(auto_flush_input())
 
+            def read_from_input_in_context() -> None:
+                # Ensure that key bindings callbacks are always executed in the
+                # current context. This is important when key bindings are
+                # accessing contextvars. (These callbacks are currently being
+                # called from a different context. Underneath,
+                # `loop.add_reader` is used to register the stdin FD.)
+                # (We copy the context to avoid a `RuntimeError` in case the
+                # context is already active.)
+                context.copy().run(read_from_input)
+
             async def auto_flush_input() -> None:
                 # Flush input after timeout.
                 # (Used for flushing the enter key.)
@@ -726,7 +748,7 @@ class Application(Generic[_AppResult]):
 
             # Enter raw mode, attach input and attach WINCH event handler.
             with self.input.raw_mode(), self.input.attach(
-                read_from_input
+                read_from_input_in_context
             ), attach_winch_signal_handler(self._on_resize):
                 # Draw UI.
                 self._request_absolute_cursor_position()
@@ -779,14 +801,16 @@ class Application(Generic[_AppResult]):
                 return result
 
         @contextmanager
-        def get_loop() -> Iterator[AbstractEventLoop]:
+        def set_loop() -> Iterator[AbstractEventLoop]:
             loop = get_running_loop()
             self.loop = loop
+            self._loop_thread = threading.current_thread()
 
             try:
                 yield loop
             finally:
                 self.loop = None
+                self._loop_thread = None
 
         @contextmanager
         def set_is_running() -> Iterator[None]:
@@ -799,6 +823,10 @@ class Application(Generic[_AppResult]):
         @contextmanager
         def set_handle_sigint(loop: AbstractEventLoop) -> Iterator[None]:
             if handle_sigint:
+                # save sigint handlers (python and os level)
+                # See: https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1576
+                sigint = signal.getsignal(signal.SIGINT)
+                sigint_os = pythonapi.PyOS_getsig(signal.SIGINT)
                 loop.add_signal_handler(
                     signal.SIGINT,
                     lambda *_: loop.call_soon_threadsafe(
@@ -809,6 +837,8 @@ class Application(Generic[_AppResult]):
                     yield
                 finally:
                     loop.remove_signal_handler(signal.SIGINT)
+                    signal.signal(signal.SIGINT, sigint)
+                    pythonapi.PyOS_setsig(signal.SIGINT, sigint_os)
             else:
                 yield
 
@@ -861,7 +891,7 @@ class Application(Generic[_AppResult]):
             # `max_postpone_time`.
             self._invalidated = False
 
-            loop = stack.enter_context(get_loop())
+            loop = stack.enter_context(set_loop())
 
             stack.enter_context(set_handle_sigint(loop))
             stack.enter_context(set_exception_handler_ctx(loop))
@@ -892,13 +922,12 @@ class Application(Generic[_AppResult]):
         set_exception_handler: bool = True,
         handle_sigint: bool = True,
         in_thread: bool = False,
+        inputhook: InputHook | None = None,
     ) -> _AppResult:
         """
         A blocking 'run' call that waits until the UI is finished.
 
-        This will start the current asyncio event loop. If no loop is set for
-        the current thread, then it will create a new loop. If a new loop was
-        created, this won't close the new loop (if `in_thread=False`).
+        This will run the application in a fresh asyncio event loop.
 
         :param pre_run: Optional callable, which is called right after the
             "reset" of the application.
@@ -931,6 +960,7 @@ class Application(Generic[_AppResult]):
                         set_exception_handler=set_exception_handler,
                         # Signal handling only works in the main thread.
                         handle_sigint=False,
+                        inputhook=inputhook,
                     )
                 except BaseException as e:
                     exception = e
@@ -948,17 +978,47 @@ class Application(Generic[_AppResult]):
             set_exception_handler=set_exception_handler,
             handle_sigint=handle_sigint,
         )
-        try:
-            # See whether a loop was installed already. If so, use that. That's
-            # required for the input hooks to work, they are installed using
-            # `set_event_loop`.
-            loop = asyncio.get_event_loop()
-        except RuntimeError:
+
+        def _called_from_ipython() -> bool:
+            try:
+                return (
+                    sys.modules["IPython"].version_info < (8, 18, 0, "")
+                    and "IPython/terminal/interactiveshell.py"
+                    in sys._getframe(3).f_code.co_filename
+                )
+            except BaseException:
+                return False
+
+        if inputhook is not None:
+            # Create new event loop with given input hook and run the app.
+            # In Python 3.12, we can use asyncio.run(loop_factory=...)
+            # For now, use `run_until_complete()`.
+            loop = new_eventloop_with_inputhook(inputhook)
+            result = loop.run_until_complete(coro)
+            loop.run_until_complete(loop.shutdown_asyncgens())
+            loop.close()
+            return result
+
+        elif _called_from_ipython():
+            # workaround to make input hooks work for IPython until
+            # https://github.com/ipython/ipython/pull/14241 is merged.
+            # IPython was setting the input hook by installing an event loop
+            # previously.
+            try:
+                # See whether a loop was installed already. If so, use that.
+                # That's required for the input hooks to work, they are
+                # installed using `set_event_loop`.
+                loop = asyncio.get_event_loop()
+            except RuntimeError:
+                # No loop installed. Run like usual.
+                return asyncio.run(coro)
+            else:
+                # Use existing loop.
+                return loop.run_until_complete(coro)
+
+        else:
             # No loop installed. Run like usual.
             return asyncio.run(coro)
-        else:
-            # Use existing loop.
-            return loop.run_until_complete(coro)
 
     def _handle_exception(
         self, loop: AbstractEventLoop, context: dict[str, Any]
@@ -993,7 +1053,7 @@ class Application(Generic[_AppResult]):
         manager. (We will only install the hook if no other custom hook was
         set.)
         """
-        if sys.version_info >= (3, 7) and sys.breakpointhook == sys.__breakpointhook__:
+        if sys.breakpointhook == sys.__breakpointhook__:
             sys.breakpointhook = self._breakpointhook
 
             try:
@@ -1007,6 +1067,10 @@ class Application(Generic[_AppResult]):
         """
         Breakpointhook which uses PDB, but ensures that the application is
         hidden and input echoing is restored during each debugger dispatch.
+
+        This can be called from any thread. In any case, the application's
+        event loop will be blocked while the PDB input is displayed. The event
+        will continue after leaving the debugger.
         """
         app = self
         # Inline import on purpose. We don't want to import pdb, if not needed.
@@ -1015,22 +1079,71 @@ class Application(Generic[_AppResult]):
 
         TraceDispatch = Callable[[FrameType, str, Any], Any]
 
-        class CustomPdb(pdb.Pdb):
-            def trace_dispatch(
-                self, frame: FrameType, event: str, arg: Any
-            ) -> TraceDispatch:
+        @contextmanager
+        def hide_app_from_eventloop_thread() -> Generator[None, None, None]:
+            """Stop application if `__breakpointhook__` is called from within
+            the App's event loop."""
+            # Hide application.
+            app.renderer.erase()
+
+            # Detach input and dispatch to debugger.
+            with app.input.detach():
+                with app.input.cooked_mode():
+                    yield
+
+            # Note: we don't render the application again here, because
+            # there's a good chance that there's a breakpoint on the next
+            # line. This paint/erase cycle would move the PDB prompt back
+            # to the middle of the screen.
+
+        @contextmanager
+        def hide_app_from_other_thread() -> Generator[None, None, None]:
+            """Stop application if `__breakpointhook__` is called from a
+            thread other than the App's event loop."""
+            ready = threading.Event()
+            done = threading.Event()
+
+            async def in_loop() -> None:
+                # from .run_in_terminal import in_terminal
+                # async with in_terminal():
+                #     ready.set()
+                #     await asyncio.get_running_loop().run_in_executor(None, done.wait)
+                #     return
+
                 # Hide application.
                 app.renderer.erase()
 
                 # Detach input and dispatch to debugger.
                 with app.input.detach():
                     with app.input.cooked_mode():
+                        ready.set()
+                        # Here we block the App's event loop thread until the
+                        # debugger resumes. We could have used `with
+                        # run_in_terminal.in_terminal():` like the commented
+                        # code above, but it seems to work better if we
+                        # completely stop the main event loop while debugging.
+                        done.wait()
+
+            self.create_background_task(in_loop())
+            ready.wait()
+            try:
+                yield
+            finally:
+                done.set()
+
+        class CustomPdb(pdb.Pdb):
+            def trace_dispatch(
+                self, frame: FrameType, event: str, arg: Any
+            ) -> TraceDispatch:
+                if app._loop_thread is None:
+                    return super().trace_dispatch(frame, event, arg)
+
+                if app._loop_thread == threading.current_thread():
+                    with hide_app_from_eventloop_thread():
                         return super().trace_dispatch(frame, event, arg)
 
-                # Note: we don't render the application again here, because
-                # there's a good chance that there's a breakpoint on the next
-                # line. This paint/erase cycle would move the PDB prompt back
-                # to the middle of the screen.
+                with hide_app_from_other_thread():
+                    return super().trace_dispatch(frame, event, arg)
 
         frame = sys._getframe().f_back
         CustomPdb(stdout=sys.__stdout__).set_trace(frame)
@@ -1455,7 +1568,7 @@ async def _do_wait_for_enter(wait_text: AnyFormattedText) -> None:
 
 @contextmanager
 def attach_winch_signal_handler(
-    handler: Callable[[], None]
+    handler: Callable[[], None],
 ) -> Generator[None, None, None]:
     """
     Attach the given callback as a WINCH signal handler within the context
