@@ -1,6 +1,6 @@
 from __future__ import annotations
 
-from typing import Any, Optional
+from typing import Any
 
 from prompt_toolkit.application.current import get_app
 from prompt_toolkit.buffer import Buffer
